# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""Functions for interfacing with the neurobank repository """
import concurrent.futures
import logging
from typing import Tuple, Dict, Union, Iterator, Sequence
from urllib.parse import urlparse
from pathlib import Path

from httpx import Client, NetRCAuth, HTTPStatusError
from nbank import registry, util
from nbank.archive import resolve_extension
from nbank.core import deposit
from nbank.script import log_error

from dlab import cache

log = logging.getLogger(__name__)
default_registry = registry.default_registry()
default_auth = NetRCAuth(None)

MaybeResourcePath = Tuple[str, Union[Path, FileNotFoundError]]


def find_resources(
    *resource_ids: str,
    registry_url: str = default_registry,
    alt_base: Union[Path, str, None] = None,
    no_download: bool = False,
) -> Iterator[MaybeResourcePath]:
    """Locate resources using neurobank.

    This function will try to locate resources from the following locations:
    - a local directory (alt_base, if set),
    - a local neurobank archive (using alt_base, if provided),
    - a local cache
    - a remote HTTP archive (caching the file for local access later)

    Yields results as they become available. The result for each requested id is
    a Path if the resource was successfully located, or exist, or a
    FileNotFoundError if the resource cannot be located.

    """
    to_locate = set(resource_ids)
    if alt_base is not None:
        for resource_id in resource_ids:
            stem = Path(alt_base) / resource_id
            try:
                yield (resource_id, resolve_extension(stem))
                to_locate.remove(resource_id)
                log.debug("%s: found in alt_base", resource_id)
            except FileNotFoundError:
                pass
    if len(to_locate) == 0:
        return
    url, query = registry.get_locations_bulk(registry_url, to_locate)
    with Client() as client, concurrent.futures.ThreadPoolExecutor() as executor:
        response = util.query_registry_bulk(client, url, query)
        future_to_name = {
            executor.submit(
                fetch_resource,
                client,
                resource["locations"],
                alt_base=alt_base,
                no_download=no_download,
            ): resource["name"]
            for resource in response
        }
        for future in concurrent.futures.as_completed(future_to_name):
            resource_id = future_to_name[future]
            try:
                result = future.result()
            except FileNotFoundError as err:
                result = err
            to_locate.remove(resource_id)
            yield (resource_id, result)
    for resource_id in to_locate:
        yield (resource_id, FileNotFoundError("no such resource"))


def find_resource(
    resource_id: str,
    *,
    registry_url: str = default_registry,
    alt_base: Union[Path, str, None] = None,
    no_download: bool = False,
) -> Path:
    """Locate a resource using neurobank. This is a convenience wrapper for find_resources"""
    for _, result in find_resources(
        resource_id,
        registry_url=registry_url,
        alt_base=alt_base,
        no_download=no_download,
    ):
        if isinstance(result, Path):
            return result
        else:
            raise result


def fetch_resource(
    client: Client,
    locations: Sequence[Dict],
    *,
    alt_base: Union[Path, str, None] = None,
    no_download: bool = False,
) -> Path:
    """Fetch a resource.

    Given a sequence of locations where a resource could be found, this function
    will try to locate the file, prioritizing local archives.

    """
    # search for local files
    for loc in locations:
        if loc["scheme"] not in registry._local_schemes:
            continue
        stem = util.parse_location(loc, alt_base)
        try:
            target = resolve_extension(stem)
            log.debug("%s: found in local repository", loc["resource_name"])
            return target
        except FileNotFoundError:
            pass
    for loc in locations:
        if loc["scheme"] in registry._local_schemes:
            continue
        resource_id = loc["resource_name"]
        url = util.parse_location(loc)
        target = cache.locate(resource_id, Path(urlparse(url).netloc) / "resources")
        if target.exists():
            log.debug("%s: found in local cache", resource_id)
            return target
        if no_download:
            continue
        log.debug("%s: fetching from registry", resource_id)
        with client.stream("GET", url) as response:
            if response.status_code != 200:
                log.warn(
                    "%s: not available at %s (http status %d)",
                    resource_id,
                    url,
                    response.status_code,
                )
                continue
            with target.open("wb") as fp:
                for data in response.iter_bytes():
                    fp.write(data)
            return target
    raise FileNotFoundError(
        "resource not found in local archive, cache, or downloadable remote"
    )


def add_registry_argument(parser, dest="registry_url"):
    """Add a registry argument to an argparse parser"""
    parser.add_argument(
        "-r",
        "--registry",
        dest=dest,
        help="URL of the registry service. "
        "Default is to use the environment variable '%s'" % registry._env_registry,
        default=default_registry,
    )


def main(argv=None):
    import argparse
    from dlab.util import setup_log

    p = argparse.ArgumentParser(description="locate neurobank resources ")
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    add_registry_argument(p)
    p.add_argument(
        "-b",
        "--base",
        type=Path,
        help="set an alternative base directory to search for local resources",
    )
    p.add_argument(
        "--clear-cache",
        action="store_true",
        help="clear the contents of the local cache",
    )
    p.add_argument("id", help="identifier(s) of the resource(s) to locate", nargs="*")
    args = p.parse_args(argv)
    setup_log(args.debug)

    if args.clear_cache:
        cache.clear(urlparse(args.registry_url).netloc)

    if len(args.id) > 0:
        for name, location in find_resources(
            *args.id, registry_url=args.registry_url, alt_base=args.base
        ):
            logging.info("%s: %s", name, location)


__all__ = [
    "deposit",
    "find_resources",
    "find_resource",
    "fetch_resource",
    "add_registry_argument",
    "default_registry",
    "default_auth",
    "log_error",
    "HTTPStatusError",
]

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Signal processing functions """
import dataclasses
from typing import Tuple, Optional

import numpy as np


@dataclasses.dataclass
class Signal:
    signal: np.ndarray
    sampling_rate: float  # in Hz
    name: Optional[str] = None
    duration: float = dataclasses.field(init=False)  # in s
    dBFS: float = dataclasses.field(init=False)

    def __post_init__(self):
        self.duration = 1.0 * self.signal.size / self.sampling_rate
        self.dBFS = dBFS(self.signal)


def kernel(name: str, bandwidth: float, dt: float) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generate a smoothing kernel function with bandwidth.

    name:      the name of the kernel. can be anything supported
               by scipy.signal.get_window, plus the following functions:
        epanech, biweight, triweight, cosinus, exponential

    bandwidth: the bandwidth of the kernel, in units of dt
    dt:        the time resolution of the kernel

    Returns:
    window:    the window function, normalized such that sum(w)*dt = 1.0
    grid:      the time support of the window, centered around 0.0

    From matlab code by Zhiyi Chi
    """
    if name in ("normal", "gaussian"):
        D = 3.75 * bandwidth
    elif name == "exponential":
        D = 4.75 * bandwidth
    else:
        D = bandwidth

    N = int(np.floor(D / dt))  # number of grid points in half the support
    G = (np.arange(1, 2 * N + 2) - 1 - N) * dt  # grid support
    xv = G / bandwidth

    if name in ("gaussian", "normal"):
        W = np.exp(-xv * xv / 2)
    elif name == "exponential":
        xv = np.minimum(xv, 0)
        W = np.absolute(xv) * np.exp(xv)
    elif name in ("biweight", "quartic"):
        W = np.maximum(0, 1 - xv * xv) ** 2
    elif name == "triweight":
        W = np.maximum(0, 1 - xv * xv) ** 3
    elif name == "cosinus":
        W = np.cos(np.pi * xv / 2) * (np.absolute(xv) <= 1)
    elif name == "epanech":
        W = np.maximum(0, 1 - xv * xv)
    elif name == "hanning":
        W = np.hanning(1 + 2 * N)
    else:
        from scipy.signal import get_window

        W = get_window(name, 1 + 2 * N)

    return W / (W.sum() * dt), G


def dBFS(signal: np.ndarray) -> float:
    """Returns the RMS level of signal, in dB FS"""
    rms = np.sqrt(np.mean(signal * signal))
    return 20 * np.log10(rms) + 3.0103


def peak(signal: np.ndarray) -> float:
    """Returns the peak level of signal, in dB FS"""
    absmax = np.amax(np.absolute(signal))
    return 20 * np.log10(absmax)


def resample(signal: Signal, target: float) -> Signal:
    """Resample the signal to target rate (in Hz)"""
    import samplerate

    if signal.sampling_rate == target:
        return signal
    ratio = 1.0 * target / signal.sampling_rate
    # NB: this silently converts data to float32
    data = samplerate.resample(signal.signal, ratio, "sinc_best")
    return Signal(
        name=signal.name,
        signal=data,
        sampling_rate=target,
    )


def hp_filter(signal: Signal, cutoff: float, order: int = 12) -> Signal:
    """Highpass filter the signal to remove DC and low-frequency noise"""
    import scipy.signal as sg

    sos = sg.butter(
        order, cutoff, fs=signal.sampling_rate, btype="highpass", output="sos"
    )
    filtered = sg.sosfilt(sos, signal.signal)
    return Signal(
        name=signal.name,
        signal=filtered,
        sampling_rate=signal.sampling_rate,
    )


def rescale(signal: Signal, target: float) -> Signal:
    """Rescale the signal to a target dBFS"""
    scale = 10 ** ((target - signal.dBFS) / 20)
    return Signal(
        name=signal.name,
        signal=signal.signal * scale,
        sampling_rate=signal.sampling_rate,
    )


def ramp_signal(signal: Signal, ramp: float = 0.002) -> Signal:
    """Apply a squared cosine ramp to the beginning and end of signal."""
    from numpy import linspace, pi, sin, cos

    s = signal.signal.copy()
    n = int(ramp * signal.sampling_rate)
    t = linspace(0, pi / 2, n)
    s[:n] *= sin(t) ** 2
    s[-n:] *= cos(t) ** 2
    return Signal(
        name=signal.name,
        signal=s,
        sampling_rate=signal.sampling_rate,
    )
